@@ -18,6 +18,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/* #define this if you want the program to crash when a file system gets
+ * finalized while async handles are still outstanding.
+ */
+#undef HANDLE_ME_HARDER
+
 #include <config.h>
 
 #include "gtkfilesystem.h"
@@ -68,6 +73,11 @@ struct _GtkFileSystemUnix
   struct stat afs_statbuf;
   struct stat net_statbuf;
 
+  GHashTable *handles;
+
+  guint execute_callbacks_idle_id;
+  GSList *callbacks;
+
   guint have_afs : 1;
   guint have_net : 1;
 };
@@ -133,6 +143,7 @@ static const GtkFileInfoType STAT_NEEDED_MASK = (GTK_FILE_INFO_IS_FOLDER |
 						 GTK_FILE_INFO_ICON);
 
 static void gtk_file_system_unix_iface_init   (GtkFileSystemIface     *iface);
+static void gtk_file_system_unix_dispose      (GObject                *object);
 static void gtk_file_system_unix_finalize     (GObject                *object);
 
 static GSList *             gtk_file_system_unix_list_volumes        (GtkFileSystem     *file_system);
@@ -243,6 +254,8 @@ static GtkFileInfo *create_file_info              (GtkFileFolderUnix *folder_uni
 						   struct stat *statbuf,
 						   const char *mime_type);
 
+static gboolean execute_callbacks_idle (gpointer data);
+
 static gboolean fill_in_names     (GtkFileFolderUnix  *folder_unix,
 				   GError            **error);
 static void     fill_in_stats     (GtkFileFolderUnix  *folder_unix);
@@ -263,14 +276,14 @@ static char *       get_parent_dir    (const char       *filename);
  */
 G_DEFINE_TYPE_WITH_CODE (GtkFileSystemUnix, gtk_file_system_unix, G_TYPE_OBJECT,
 			 G_IMPLEMENT_INTERFACE (GTK_TYPE_FILE_SYSTEM,
-						gtk_file_system_unix_iface_init));
+						gtk_file_system_unix_iface_init))
 
 /*
  * GtkFileFolderUnix
  */
 G_DEFINE_TYPE_WITH_CODE (GtkFileFolderUnix, _gtk_file_folder_unix, G_TYPE_OBJECT,
 			 G_IMPLEMENT_INTERFACE (GTK_TYPE_FILE_FOLDER,
-						gtk_file_folder_unix_iface_init));
+						gtk_file_folder_unix_iface_init))
 
 
 /**
@@ -293,6 +306,7 @@ gtk_file_system_unix_class_init (GtkFileSystemUnixClass *class)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (class);
 
+  gobject_class->dispose = gtk_file_system_unix_dispose;
   gobject_class->finalize = gtk_file_system_unix_finalize;
 }
 
@@ -339,6 +353,100 @@ gtk_file_system_unix_init (GtkFileSystemUnix *system_unix)
     system_unix->have_net = TRUE;
   else
     system_unix->have_net = FALSE;
+
+  system_unix->handles = g_hash_table_new (g_direct_hash, g_direct_equal);
+
+  system_unix->execute_callbacks_idle_id = 0;
+  system_unix->callbacks = NULL;
+}
+
+static void
+check_handle_fn (gpointer key, gpointer value, gpointer data)
+{
+  GtkFileSystemHandle *handle;
+  int *num_live_handles;
+
+  handle = key;
+  num_live_handles = data;
+
+  (*num_live_handles)++;
+
+  g_warning ("file_system_unix=%p still has handle=%p at finalization which is %s!",
+	     handle->file_system,
+	     handle,
+	     handle->cancelled ? "CANCELLED" : "NOT CANCELLED");
+}
+
+static void
+check_handles_at_finalization (GtkFileSystemUnix *system_unix)
+{
+  int num_live_handles;
+
+  num_live_handles = 0;
+
+  g_hash_table_foreach (system_unix->handles, check_handle_fn, &num_live_handles);
+#ifdef HANDLE_ME_HARDER
+  g_assert (num_live_handles == 0);
+#endif
+
+  g_hash_table_destroy (system_unix->handles);
+  system_unix->handles = NULL;
+}
+
+#define GTK_TYPE_FILE_SYSTEM_HANDLE_UNIX (_gtk_file_system_handle_unix_get_type ())
+
+typedef struct _GtkFileSystemHandle GtkFileSystemHandleUnix;
+typedef struct _GtkFileSystemHandleClass GtkFileSystemHandleUnixClass;
+
+G_DEFINE_TYPE (GtkFileSystemHandleUnix, _gtk_file_system_handle_unix, GTK_TYPE_FILE_SYSTEM_HANDLE)
+
+static void
+_gtk_file_system_handle_unix_init (GtkFileSystemHandleUnix *handle)
+{
+}
+
+static void 
+_gtk_file_system_handle_unix_finalize (GObject *object)
+{
+  GtkFileSystemHandleUnix *handle;
+  GtkFileSystemUnix *system_unix;
+
+  handle = (GtkFileSystemHandleUnix *)object;
+
+  system_unix = GTK_FILE_SYSTEM_UNIX (GTK_FILE_SYSTEM_HANDLE (handle)->file_system);
+
+  g_assert (g_hash_table_lookup (system_unix->handles, handle) != NULL);
+  g_hash_table_remove (system_unix->handles, handle);
+
+  if (G_OBJECT_CLASS (_gtk_file_system_handle_unix_parent_class)->finalize)
+    G_OBJECT_CLASS (_gtk_file_system_handle_unix_parent_class)->finalize (object);
+}
+
+static void
+_gtk_file_system_handle_unix_class_init (GtkFileSystemHandleUnixClass *class)
+{
+  GObjectClass *gobject_class = (GObjectClass *) class;
+
+  gobject_class->finalize = _gtk_file_system_handle_unix_finalize;
+}
+
+static void
+gtk_file_system_unix_dispose (GObject *object)
+{
+  GtkFileSystemUnix *system_unix;
+
+  system_unix = GTK_FILE_SYSTEM_UNIX (object);
+
+  if (system_unix->execute_callbacks_idle_id)
+    {
+      g_source_remove (system_unix->execute_callbacks_idle_id);
+      system_unix->execute_callbacks_idle_id = 0;
+
+      /* call pending callbacks */
+      execute_callbacks_idle (system_unix);
+    }
+
+  G_OBJECT_CLASS (gtk_file_system_unix_parent_class)->dispose (object);
 }
 
 static void
@@ -348,6 +456,8 @@ gtk_file_system_unix_finalize (GObject *object)
 
   system_unix = GTK_FILE_SYSTEM_UNIX (object);
 
+  check_handles_at_finalization  (system_unix);
+
   /* FIXME: assert that the hash is empty? */
   g_hash_table_destroy (system_unix->folder_hash);
 
@@ -398,7 +508,7 @@ enum callback_types
   CALLBACK_VOLUME_MOUNT
 };
 
-static void queue_callback (enum callback_types type, gpointer data);
+static void queue_callback (GtkFileSystemUnix *system_unix, enum callback_types type, gpointer data);
 
 struct get_info_callback
 {
@@ -441,7 +551,7 @@ queue_get_info_callback (GtkFileSystemGetInfoCallback  callback,
   info->error = error;
   info->data = data;
 
-  queue_callback (CALLBACK_GET_INFO, info);
+  queue_callback (GTK_FILE_SYSTEM_UNIX (handle->file_system), CALLBACK_GET_INFO, info);
 }
 
 
@@ -483,7 +593,7 @@ queue_get_folder_callback (GtkFileSystemGetFolderCallback  callback,
   info->error = error;
   info->data = data;
 
-  queue_callback (CALLBACK_GET_FOLDER, info);
+  queue_callback (GTK_FILE_SYSTEM_UNIX (handle->file_system), CALLBACK_GET_FOLDER, info);
 }
 
 
@@ -528,7 +638,7 @@ queue_create_folder_callback (GtkFileSystemCreateFolderCallback  callback,
   info->error = error;
   info->data = data;
 
-  queue_callback (CALLBACK_CREATE_FOLDER, info);
+  queue_callback (GTK_FILE_SYSTEM_UNIX (handle->file_system), CALLBACK_CREATE_FOLDER, info);
 }
 
 
@@ -570,7 +680,7 @@ queue_volume_mount_callback (GtkFileSystemVolumeMountCallback  callback,
   info->error = error;
   info->data = data;
 
-  queue_callback (CALLBACK_VOLUME_MOUNT, info);
+  queue_callback (GTK_FILE_SYSTEM_UNIX (handle->file_system), CALLBACK_VOLUME_MOUNT, info);
 }
 
 
@@ -588,17 +698,22 @@ struct callback_info
 };
 
 
-static guint execute_callbacks_idle_id = 0;
-static GSList *callbacks = NULL;
 
 static gboolean
 execute_callbacks_idle (gpointer data)
 {
   GSList *l;
+  gboolean unref_file_system = TRUE;
+  GtkFileSystemUnix *system_unix = GTK_FILE_SYSTEM_UNIX (data);
 
   GDK_THREADS_ENTER ();
 
-  for (l = callbacks; l; l = l->next)
+  if (!system_unix->execute_callbacks_idle_id)
+    unref_file_system = FALSE;
+  else
+    g_object_ref (system_unix);
+
+  for (l = system_unix->callbacks; l; l = l->next)
     {
       struct callback_info *info = l->data;
 
@@ -624,10 +739,13 @@ execute_callbacks_idle (gpointer data)
       g_free (info);
     }
 
-  g_slist_free (callbacks);
-  callbacks = NULL;
+  g_slist_free (system_unix->callbacks);
+  system_unix->callbacks = NULL;
 
-  execute_callbacks_idle_id = 0;
+  if (unref_file_system)
+    g_object_unref (system_unix);
+
+  system_unix->execute_callbacks_idle_id = 0;
 
   GDK_THREADS_LEAVE ();
 
@@ -635,7 +753,9 @@ execute_callbacks_idle (gpointer data)
 }
 
 static void
-queue_callback (enum callback_types type, gpointer data)
+queue_callback (GtkFileSystemUnix   *system_unix,
+		enum callback_types  type,
+		gpointer             data)
 {
   struct callback_info *info;
 
@@ -661,20 +781,26 @@ queue_callback (enum callback_types type, gpointer data)
 	break;
     }
 
-  callbacks = g_slist_append (callbacks, info);
+  system_unix->callbacks = g_slist_append (system_unix->callbacks, info);
 
-  if (!execute_callbacks_idle_id)
-    execute_callbacks_idle_id = g_idle_add (execute_callbacks_idle, NULL);
+  if (!system_unix->execute_callbacks_idle_id)
+    system_unix->execute_callbacks_idle_id = g_idle_add (execute_callbacks_idle, system_unix);
 }
 
 static GtkFileSystemHandle *
 create_handle (GtkFileSystem *file_system)
 {
+  GtkFileSystemUnix *system_unix;
   GtkFileSystemHandle *handle;
 
-  handle = g_object_new (GTK_TYPE_FILE_SYSTEM_HANDLE, NULL);
+  system_unix = GTK_FILE_SYSTEM_UNIX (file_system);
+
+  handle = g_object_new (GTK_TYPE_FILE_SYSTEM_HANDLE_UNIX, NULL);
   handle->file_system = file_system;
 
+  g_assert (g_hash_table_lookup (system_unix->handles, handle) == NULL);
+  g_hash_table_insert (system_unix->handles, handle, handle);
+
   return handle;
 }
 
@@ -885,8 +1011,9 @@ gtk_file_system_unix_get_folder (GtkFileSystem                  *file_system,
   queue_get_folder_callback (callback, handle, GTK_FILE_FOLDER (folder_unix), NULL, data);
 
   /* Start loading the folder contents in an idle */
-  folder_unix->load_folder_id =
-    g_idle_add ((GSourceFunc) load_folder, folder_unix);
+  if (!folder_unix->load_folder_id)
+    folder_unix->load_folder_id =
+      g_idle_add ((GSourceFunc) load_folder, folder_unix);
 
   return handle;
 }
@@ -2074,8 +2201,8 @@ file_info_for_root_with_error (const char  *root_name,
       g_set_error (error,
 		   GTK_FILE_SYSTEM_ERROR,
 		   GTK_FILE_SYSTEM_ERROR_FAILED,
-		   _("Error getting information for '/': %s"),
-		   g_strerror (saved_errno));
+		   _("Error getting information for '%s': %s"),
+		   "/", g_strerror (saved_errno));
 
       return NULL;
     }
